/**
 * - [ ] TODO: look into adding back the `groupScalars` config option, that way
 *       the generated schemas are more likely to be "deeper" without risk of stack overflow
 */
import type * as T from '@traversable/registry';
import { URI } from '@traversable/registry';
import { Json } from '@traversable/json';
import type { SchemaOptions } from '@traversable/schema';
import { t } from '@traversable/schema';
import * as fc from './fast-check.js';
export { type Arbitraries, type Builder, type Constraints, type Fixpoint, type Free, type Nullary, type Positional, type Seed, type SortBias, type SpecialCase, type TypeName, type Unary, integerF as integer, bigintF as bigint, numberF as number, stringF as string, eqF as eq, optionalF as optional, arrayF as array, recordF as record, unionF as union, intersectF as intersect, objectF as object, tupleF as tuple, data, defaults, defineSeed, extensibleArbitrary, fold, fromJsonLiteral, fromSchema, Functor, getBounds, identity, initialOrder, is, isBoundable, isBoundableTag, isAssociative, isNullary, isPositional, isSeed, isSpecialCase, isUnary, numberConstraintsFromBounds, parseConstraints, Recursive, schema, schemaWithMinDepth, seed, stringConstraintsFromBounds, toArbitrary, toSchema, toString, unfold, };
/**
 * If you provide a partial weight map, missing properties will fall back to `0`
 */
type SortBias<T> = T.Comparator<keyof T> | {
    [K in keyof T]+?: number;
};
type TypeName = Exclude<keyof Builder, 'tree'>;
declare const initialOrderMap: {
    readonly string: 0;
    readonly number: 1;
    readonly object: 2;
    readonly boolean: 3;
    readonly undefined: 4;
    readonly symbol: 5;
    readonly integer: 6;
    readonly bigint: 7;
    readonly null: 8;
    readonly eq: 9;
    readonly array: 9;
    readonly record: 10;
    readonly optional: 11;
    readonly tuple: 12;
    readonly intersect: 13;
    readonly union: 14;
    readonly any: 15;
    readonly unknown: 16;
    readonly void: 17;
    readonly never: 18;
};
declare const initialOrder: (keyof typeof initialOrderMap)[];
type LibConstraints<Exclude extends TypeName, Include extends TypeName = TypeName> = {
    sortBias?: SortBias<Builder>;
    exclude?: Exclude[];
    include?: Include[];
};
type ObjectConstraints<T, U> = {
    min?: number;
    max?: number;
} & Omit<TargetConstraints<T, U>['object'], 'minLength' | 'maxLength'>;
type Arbitraries = {
    never?: unknown;
    unknown?: unknown;
    void?: unknown;
    any?: unknown;
    undefined?: unknown;
    null?: unknown;
    symbol?: unknown;
    boolean?: unknown;
    integer?: unknown;
    number?: unknown;
    bigint?: unknown;
    string?: unknown;
    eq?(x: unknown, $?: SchemaOptions): unknown;
    array?(x: unknown, $?: SchemaOptions): unknown;
    record?(x: unknown, $?: SchemaOptions): unknown;
    optional?(x: unknown, $?: SchemaOptions): unknown;
    union?(x: readonly unknown[], $?: SchemaOptions): unknown;
    intersect?(x: readonly unknown[], $?: SchemaOptions): unknown;
    tuple?(x: readonly unknown[], $?: SchemaOptions): unknown;
    object?(x: {
        [x: string]: unknown;
    }, $?: SchemaOptions): unknown;
};
type Constraints<Exclude extends TypeName, Include extends TypeName = TypeName, T = unknown, U = T> = LibConstraints<Exclude, Include> & {
    arbitraries?: Arbitraries;
    integer?: TargetConstraints['integer'];
    bigint?: TargetConstraints['bigint'];
    number?: TargetConstraints['number'];
    string?: TargetConstraints['string'];
    array?: TargetConstraints['array'];
    union?: TargetConstraints['union'];
    intersect?: TargetConstraints['intersect'];
    tree?: TargetConstraints['tree'];
    object?: ObjectConstraints<T, U>;
    tuple?: TargetConstraints['tuple'];
};
declare const defaults: {
    arbitraries: {};
    integer: {
        min: number;
        max: number;
    };
    bigint: {
        min: bigint;
        max: bigint;
    };
    number: {
        min: number;
        max: number;
        minExcluded: false;
        maxExcluded: false;
        noNaN: true;
    };
    string: {
        min: number;
        max: number;
    };
    array: {
        min: number;
        max: number;
    };
    union: {
        readonly minLength: 2;
        readonly maxLength: 2;
        readonly size: "xsmall";
    };
    intersect: {
        readonly minLength: 1;
        readonly maxLength: 2;
        readonly size: "xsmall";
        readonly depthIdentifier: fc.DepthContext;
    };
    tuple: {
        readonly minLength: 1;
        readonly maxLength: 3;
        readonly size: "xsmall";
        readonly depthIdentifier: fc.DepthContext;
    };
    object: {
        min: number;
        max: number;
        size: "xsmall";
    };
    tree: {
        readonly maxDepth: 3;
        readonly depthIdentifier: fc.DepthContext;
        readonly depthSize: "xsmall";
        readonly withCrossShrink: false;
    };
    sortBias: () => number;
    include: ("string" | "number" | "bigint" | "boolean" | "symbol" | "undefined" | "object" | "never" | "any" | "unknown" | "void" | "null" | "integer" | "eq" | "array" | "record" | "optional" | "tuple" | "union" | "intersect")[];
    exclude: [];
};
interface InclusiveBounds<T = number> {
    minimum?: T;
    maximum?: T;
}
interface ExclusiveBounds<T = number> {
    exclusiveMinimum?: T;
    exclusiveMaximum?: T;
}
interface StringBounds extends InclusiveBounds {
}
interface ArrayBounds extends InclusiveBounds {
}
interface IntegerBounds extends InclusiveBounds {
}
interface NumberBounds extends InclusiveBounds, ExclusiveBounds {
}
interface BigIntBounds extends InclusiveBounds<bigint | number> {
}
interface integerF extends T.inline<[tag: URI.integer, constraints?: IntegerBounds]> {
}
interface bigintF extends T.inline<[tag: URI.bigint, constraints?: BigIntBounds]> {
}
interface numberF extends T.inline<[tag: URI.number, constraints?: NumberBounds]> {
}
interface stringF extends T.inline<[tag: URI.string, constraints?: StringBounds]> {
}
interface arrayF<S> extends T.inline<[tag: URI.array, def: S, constraints?: ArrayBounds]> {
}
declare function integerF(constraints?: IntegerBounds): integerF;
declare function bigintF(constraints?: BigIntBounds): bigintF;
declare function numberF(constraints?: NumberBounds): numberF;
declare function stringF(constraints?: StringBounds): stringF;
declare function arrayF<S>(def: S, constraints?: ArrayBounds): arrayF<S>;
interface eqF<S = Json> extends T.inline<[tag: URI.eq, def: S]> {
}
interface optionalF<S> extends T.inline<[tag: URI.optional, def: S]> {
}
interface recordF<S> extends T.inline<[tag: URI.record, def: S]> {
}
interface objectF<S> extends T.inline<[tag: URI.object, def: S]> {
}
interface tupleF<S> extends T.inline<[tag: URI.tuple, def: S]> {
}
interface unionF<S> extends T.inline<[tag: URI.union, def: S]> {
}
interface intersectF<S> extends T.inline<[tag: URI.intersect, def: S]> {
}
declare function eqF<V = Json>(def: V): eqF<V>;
declare function optionalF<S>(def: S): optionalF<S>;
declare function recordF<S>(def: S): recordF<S>;
declare function objectF<S extends [k: string, v: unknown][]>(def: readonly [...S]): objectF<[...S]>;
declare function tupleF<S extends readonly unknown[]>(def: readonly [...S]): tupleF<readonly [...S]>;
declare function unionF<S extends readonly unknown[]>(def: readonly [...S]): unionF<readonly [...S]>;
declare function intersectF<S extends readonly unknown[]>(def: readonly [...S]): intersectF<readonly [...S]>;
type Seed<F> = Nullary | Boundable | eqF<F> | arrayF<F> | recordF<F> | optionalF<F> | tupleF<readonly F[]> | unionF<readonly F[]> | intersectF<readonly F[]> | objectF<[k: string, v: F][]>;
type Fixpoint = Nullary | Boundable | eqF | arrayF<Fixpoint> | recordF<Fixpoint> | optionalF<Fixpoint> | tupleF<readonly Fixpoint[]> | unionF<readonly Fixpoint[]> | intersectF<readonly Fixpoint[]> | objectF<[k: string, v: Fixpoint][]>;
interface Free extends T.HKT {
    [-1]: Seed<this[0]>;
}
declare namespace Seed {
    export { Builder, Free, Fixpoint, Inductive, Nullary, };
}
type Nullary = typeof NullaryTags[number];
declare const NullaryTags: ["@traversable/schema/URI::never", "@traversable/schema/URI::any", "@traversable/schema/URI::unknown", "@traversable/schema/URI::void", "@traversable/schema/URI::undefined", "@traversable/schema/URI::null", "@traversable/schema/URI::boolean", "@traversable/schema/URI::symbol"];
declare const isNullary: (u: unknown) => u is Nullary;
export declare const laxMin: (...xs: (number | undefined)[]) => number | undefined;
export declare const laxMax: (...xs: (number | undefined)[]) => number | undefined;
/**
 * It's _crazy_ how complicated this is.
 *
 * Surely there's a more elegant way to accomplish this.
 */
export declare const preprocessNumberBounds: ({ minimum: min_, maximum: max_, exclusiveMinimum: xMin_, exclusiveMaximum: xMax_ }: NumberBounds) => NumberBounds;
type Boundable = integerF | bigintF | numberF | stringF;
type BoundableTag = typeof BoundableTags[number];
declare const BoundableTags: ["@traversable/schema/URI::bigint", "@traversable/schema/URI::integer", "@traversable/schema/URI::number", "@traversable/schema/URI::string"];
declare const isBoundableTag: (u: unknown) => u is BoundableTag;
declare const isBoundable: (u: unknown) => u is Boundable;
type SpecialCase<T = unknown> = [SpecialCaseTag, T];
type SpecialCaseTag = typeof SpecialCaseTags[number];
declare const SpecialCaseTags: ["@traversable/schema/URI::eq"];
declare const isSpecialCase: <T>(u: unknown) => u is SpecialCase<T>;
type Unary<T = unknown> = arrayF<T> | recordF<T> | optionalF<T>;
declare const isUnary: <T>(u: unknown) => u is Unary<T>;
type Positional<T = unknown> = tupleF<T> | unionF<T> | intersectF<T>;
declare const isPositional: <T>(u: unknown) => u is Positional<T>;
type Associative<T = unknown> = [AssociativeTag, [k: string, v: T][]];
type AssociativeTag = typeof AssociativeTags[number];
declare const AssociativeTags: ["@traversable/schema/URI::object"];
declare const isAssociative: <T>(u: unknown) => u is Associative<T>;
declare const isSeed: (u: unknown) => u is unknown;
/**
 * ## {@link is `is`}
 *
 * Type-guard from `unknown` to {@link Seed `Seed`}.
 *
 * The {@link is `is`} function is also an object whose properties narrow
 * to a particular member or subset of members of {@link Seed `Seed`}.
 */
declare function is(u: unknown): boolean;
declare namespace is {
    export var nullary: (u: unknown) => u is Nullary;
    export var unary: <T>(u: unknown) => u is Unary<T>;
    export var positional: <T>(u: unknown) => u is Positional<T>;
    export var associative: <T>(u: unknown) => u is Associative<T>;
    export var special: <T>(u: unknown) => u is SpecialCase<T>;
    export var never: (u: unknown) => u is "@traversable/schema/URI::never";
    export var any: (u: unknown) => u is "@traversable/schema/URI::any";
    export var unknown: (u: unknown) => u is "@traversable/schema/URI::unknown";
    var _a: (u: unknown) => u is "@traversable/schema/URI::void";
    var _b: (u: unknown) => u is "@traversable/schema/URI::null";
    export var undefined: (u: unknown) => u is "@traversable/schema/URI::undefined";
    export var boolean: (u: unknown) => u is "@traversable/schema/URI::boolean";
    export var symbol: (u: unknown) => u is "@traversable/schema/URI::symbol";
    export var string: t.has<[0], t.of<(u: unknown) => u is "@traversable/schema/URI::string">>;
    export var number: t.has<[0], t.of<(u: unknown) => u is "@traversable/schema/URI::number">>;
    export var integer: t.has<[0], t.of<(u: unknown) => u is "@traversable/schema/URI::integer">>;
    export var bigint: t.has<[0], t.of<(u: unknown) => u is "@traversable/schema/URI::bigint">>;
    export var eq: <T>(u: unknown) => u is [tag: URI.eq, def: T];
    export var array: <T>(u: unknown) => u is [tag: URI.array, T];
    export var optional: <T>(u: unknown) => u is [tag: URI.optional, T];
    export var record: <T>(u: unknown) => u is [tag: URI.record, T];
    export var union: <T>(u: unknown) => u is [tag: URI.union, readonly T[]];
    export var intersect: <T>(u: unknown) => u is [tag: URI.intersect, readonly T[]];
    export var tuple: <T>(u: unknown) => u is [tag: URI.tuple, readonly T[]];
    export var object: <T>(u: unknown) => u is [tag: URI.tuple, {
        [x: string]: T;
    }];
    export { _a as void, _b as null };
}
type Inductive<S> = [S] extends [infer T extends Nullary] ? T : [S] extends [readonly [Unary, infer T]] ? [tag: S[0], unary: Inductive<T>] : [S] extends [readonly [Positional, infer T extends readonly unknown[]]] ? [S[0], {
    -readonly [Ix in keyof T]: Inductive<T[Ix]>;
}] : [S] extends [readonly [Associative, infer T extends readonly [k: string, v: unknown][]]] ? [S[0], {
    -readonly [Ix in keyof T]: [k: T[Ix][0], v: Inductive<T[Ix][1]>];
}] : T.TypeError<'Expected: Fixpoint'>;
/**
 * Hand-tuned constructor that gives you both more precise and
 * more localized feedback than the TS compiler when you make a mistake
 * in an inline {@link Seed `Seed`} definition.
 *
 * When you're working with deeply nested tuples where only certain sequences
 * are valid constructions, this turns out to be pretty useful / necessary.
 */
declare function defineSeed<T extends Inductive<T>>(seed: T): T;
interface Builder {
    never: URI.never;
    any: URI.any;
    unknown: URI.unknown;
    void: URI.void;
    null: URI.null;
    undefined: URI.undefined;
    symbol: URI.symbol;
    boolean: URI.boolean;
    bigint: [URI.bigint, BigIntBounds];
    integer: [URI.integer, IntegerBounds];
    number: [URI.number, NumberBounds];
    string: [URI.string, StringBounds];
    eq: [tag: URI.eq, seed: Json];
    array: [tag: URI.array, seed: Fixpoint, ArrayBounds];
    record: [tag: URI.record, seed: Fixpoint];
    optional: [tag: URI.optional, seed: Fixpoint];
    tuple: [tag: URI.tuple, seed: readonly Fixpoint[]];
    union: [tag: URI.union, seed: readonly Seed.Fixpoint[]];
    intersect: [tag: URI.intersect, seed: readonly Seed.Fixpoint[]];
    object: [tag: URI.object, seed: [k: string, Fixpoint][]];
    tree: Omit<this, 'tree'>[keyof Omit<this, 'tree'>];
}
declare const numberConstraintsFromBounds: (bounds?: InclusiveBounds<number> & ExclusiveBounds<number>) => fc.FloatConstraints;
declare const stringConstraintsFromBounds: ({ minimum: min, maximum: max }?: InclusiveBounds) => fc.StringConstraints;
declare const Functor: T.Functor<Seed.Free, Seed.Fixpoint>;
declare const fold: <T>(algebra: T.Algebra<Free, T>) => <S extends Fixpoint>(term: S) => T;
declare const unfold: <T>(coalgebra: T.Coalgebra<Free, T>) => <S extends Fixpoint>(expr: S) => Seed<T>;
type t_Boundable = t.Boundable | t.array<any>;
declare function getBounds(x: t_Boundable): (IntegerBounds & NumberBounds & BigIntBounds & StringBounds) | undefined;
declare namespace Recursive {
    const identity: T.Functor.Algebra<Seed.Free, Seed.Fixpoint>;
    const toExtensibleSchema: (constraints?: Constraints<never>['arbitraries']) => T.Functor.Algebra<Seed.Free, unknown>;
    const toSchema: T.Functor.Algebra<Seed.Free, t.Schema>;
    const fromSchema: T.Functor.Algebra<t.Free, Seed.Fixpoint>;
    const toArbitrary: T.Functor.Algebra<Seed.Free, fc.Arbitrary>;
    const toString: T.Functor.Algebra<Seed.Free, string>;
    const fromJsonLiteral: T.Functor.Algebra<Json.Free, Seed.Fixpoint>;
}
export declare const sortOptionalsLast: (l: unknown, r: unknown) => 0 | 1 | -1;
export declare const pickAndSortNodes: (nodes: readonly TypeName[]) => <Exclude extends TypeName, Include extends TypeName>(constraints?: Pick<TargetConstraints<_, _, Exclude, Include>, 'exclude' | 'include' | 'sortBias'>) => TypeName[];
declare function parseConstraints<Exclude extends TypeName, Include extends TypeName, T, U>(constraints?: Constraints<Exclude, Include, T, U>): Required<TargetConstraints<T, U, Exclude, Include>>;
export declare const toJson: (seed: Seed.Fixpoint) => Json.Fixpoint;
type SeedIR = {
    eq: fc.Arbitrary<eqF<fc.JsonValue>>;
    array: fc.Arbitrary<arrayF<Fixpoint>>;
    record: fc.Arbitrary<recordF<Fixpoint>>;
    optional: fc.Arbitrary<optionalF<Fixpoint>>;
    tuple: fc.Arbitrary<tupleF<readonly Fixpoint[]>>;
    object: fc.Arbitrary<objectF<[k: string, v: Fixpoint][]>>;
    union: fc.Arbitrary<unionF<readonly Fixpoint[]>>;
    intersect: fc.Arbitrary<intersectF<readonly Fixpoint[]>>;
    never: fc.Arbitrary<URI.never>;
    any: fc.Arbitrary<URI.any>;
    unknown: fc.Arbitrary<URI.unknown>;
    void: fc.Arbitrary<URI.void>;
    null: fc.Arbitrary<URI.null>;
    undefined: fc.Arbitrary<URI.undefined>;
    symbol: fc.Arbitrary<URI.symbol>;
    boolean: fc.Arbitrary<URI.boolean>;
    integer: fc.Arbitrary<[URI.integer, _?: IntegerBounds]>;
    bigint: fc.Arbitrary<[URI.bigint, _?: BigIntBounds]>;
    number: fc.Arbitrary<[URI.number, _?: NumberBounds]>;
    string: fc.Arbitrary<[URI.string, _?: StringBounds]>;
};
type SeedResult<Exclude extends TypeName, Include extends TypeName = TypeName> = Pick<SeedIR, globalThis.Exclude<Include, Exclude>> & Omit<SeedIR, Exclude> & Tree<Exclude>;
type Tree<K extends keyof SeedIR = never> = {
    tree: Omit<SeedIR, 'tree' | K>[keyof Omit<SeedIR, 'tree' | K>];
};
declare function seed<Include extends TypeName, Exclude extends TypeName = never>(_: Constraints<Exclude, Include>): (go: fc.LetrecTypedTie<Builder>) => SeedResult<Exclude, Include>;
declare function seed(): (go: fc.LetrecTypedTie<Builder>) => SeedResult<never>;
declare function seed(_?: Constraints<never>): (go: fc.LetrecTypedTie<Builder>) => SeedResult<never>;
declare function schemaWithMinDepth<Exclude extends TypeName, Include extends TypeName>(_: Constraints<Exclude, Include> | undefined, n: number): fc.Arbitrary<t.Schema>;
declare const identity: <S extends Fixpoint>(term: S) => Fixpoint;
declare const toSchema: <S extends Fixpoint>(term: S) => t.Schema<t.Unspecified>;
declare const toArbitrary: <S extends Fixpoint>(term: S) => fc.Arbitrary<unknown>;
declare const toString: <S extends Fixpoint>(term: S) => string;
declare const fromSchema: <S extends t.Schema<t.Unspecified>>(term: S) => Fixpoint;
declare const fromJsonLiteral: <S extends Fixpoint>(term: S) => Fixpoint;
/**
 * ## {@link schema `Seed.schema`}
 *
 * Generates an arbitrary,
 * [pseudo-random](https://en.wikipedia.org/wiki/Pseudorandomness)
 * {@link t `t`} schema.
 *
 * Internally, schemas are generated from a
 * {@link Seed `seed value`}, which is itself generated by a library
 * called [`fast-check`](https://github.com/dubzzz/fast-check).
 */
declare function schema<Include extends TypeName = TypeName, Exclude extends TypeName = never>(constraints?: Constraints<Exclude, Include>): fc.Arbitrary<globalThis.Exclude<t.F<Fixpoint>, {
    tag: `${T.NS}${Exclude}`;
}>>;
declare function schema(constraints?: Constraints<never>): fc.Arbitrary<t.LowerBound>;
declare const extensibleArbitrary: <T>(constraints?: Constraints<never>) => import("fast-check").Arbitrary<unknown>;
/**
 * ## {@link data `Seed.data`}
 *
 * Generates an arbitrary,
 * [pseudo-random](https://en.wikipedia.org/wiki/Pseudorandomness)
 * data generator.
 *
 * Internally, schemas are generated from a
 * {@link Seed `seed value`}, which is itself generated by a library
 * called [`fast-check`](https://github.com/dubzzz/fast-check).
 */
declare const data: (constraints?: Constraints<never>) => import("fast-check").Arbitrary<unknown>;
//# sourceMappingURL=seed.d.ts.map