/**
 * - [ ] TODO: look into adding back the `groupScalars` config option, that way
 *       the generated schemas are more likely to be "deeper" without risk of stack overflow
 */
import { fn, parseKey, unsafeCompact, URI } from '@traversable/registry';
import { Json } from '@traversable/json';
import { t } from '@traversable/schema';
import * as fc from './fast-check.js';
export { integerF as integer, bigintF as bigint, numberF as number, stringF as string, eqF as eq, optionalF as optional, arrayF as array, recordF as record, unionF as union, intersectF as intersect, objectF as object, tupleF as tuple, data, defaults, defineSeed, extensibleArbitrary, fold, fromJsonLiteral, fromSchema, Functor, getBounds, identity, initialOrder, is, isBoundable, isBoundableTag, isAssociative, isNullary, isPositional, isSeed, isSpecialCase, isUnary, numberConstraintsFromBounds, parseConstraints, Recursive, schema, schemaWithMinDepth, seed, stringConstraintsFromBounds, toArbitrary, toSchema, toString, unfold };
/** @internal */
const Object_fromEntries = globalThis.Object.fromEntries;
/** @internal */
const Object_assign = globalThis.Object.assign;
/** @internal */
const Array_isArray = globalThis.Array.isArray;
/** @internal */
const opts = {
  optionalTreatment: 'treatUndefinedAndOptionalAsTheSame'
};
/** @internal */
const isComposite = u => Array_isArray(u) || u !== null && typeof u === 'object';
/** @internal */
const isNumeric = /*#__PURE__*/t.union(t.number, t.bigint);
const initialOrderMap = {
  string: 0,
  number: 1,
  object: 2,
  boolean: 3,
  undefined: 4,
  symbol: 5,
  integer: 6,
  bigint: 7,
  null: 8,
  eq: 9,
  array: 9,
  record: 10,
  optional: 11,
  tuple: 12,
  intersect: 13,
  union: 14,
  any: 15,
  unknown: 16,
  void: 17,
  never: 18
};
const initialOrder = /*#__PURE__*/Object.entries(initialOrderMap).sort(([, l], [, r]) => l < r ? -1 : l > r ? 1 : 0).map(([k]) => k);
const defaultDepthIdentifier = /*#__PURE__*/fc.createDepthIdentifier();
const defaultTupleConstraints = {
  minLength: 1,
  maxLength: 3,
  size: 'xsmall',
  depthIdentifier: defaultDepthIdentifier
};
const defaultIntersectConstraints = {
  minLength: 1,
  maxLength: 2,
  size: 'xsmall',
  depthIdentifier: defaultDepthIdentifier
};
const defaultUnionConstraints = {
  minLength: 2,
  maxLength: 2,
  size: 'xsmall'
};
const defaultObjectConstraints = {
  min: 1,
  max: 3,
  size: 'xsmall'
};
const defaultTreeConstraints = {
  maxDepth: 3,
  depthIdentifier: defaultDepthIdentifier,
  depthSize: 'xsmall',
  withCrossShrink: false
};
const defaultIntegerConstraints = {
  min: -0x100,
  max: 0x100
};
const defaultNumberConstraints = {
  min: -0x10000,
  max: 0x10000,
  minExcluded: false,
  maxExcluded: false,
  noNaN: true
};
const defaultBigIntConstraints = {
  min: -0x1000000n,
  max: 0x1000000n
};
const defaultStringConstraints = {
  min: 0,
  max: 0x100
};
const defaultArrayConstraints = {
  min: 0,
  max: 0x100
};
const defaults = {
  arbitraries: {},
  integer: defaultIntegerConstraints,
  bigint: defaultBigIntConstraints,
  number: defaultNumberConstraints,
  string: defaultStringConstraints,
  array: defaultArrayConstraints,
  union: defaultUnionConstraints,
  intersect: defaultIntersectConstraints,
  tuple: defaultTupleConstraints,
  object: defaultObjectConstraints,
  tree: defaultTreeConstraints,
  sortBias: () => 0,
  include: initialOrder,
  exclude: []
  // groupScalars: true,
};
function integerF(constraints) {
  return !constraints ? [URI.integer] : [URI.integer, constraints];
}
function bigintF(constraints) {
  return !constraints ? [URI.bigint] : [URI.bigint, constraints];
}
function numberF(constraints) {
  return !constraints ? [URI.number] : [URI.number, constraints];
}
function stringF(constraints) {
  return !constraints ? [URI.string] : [URI.string, constraints];
}
function arrayF(def, constraints) {
  return !constraints ? [URI.array, def] : [URI.array, def, constraints];
}
function eqF(def) {
  return [URI.eq, def];
}
function optionalF(def) {
  return [URI.optional, def];
}
function recordF(def) {
  return [URI.record, def];
}
function objectF(def) {
  return [URI.object, [...def]];
}
function tupleF(def) {
  return [URI.tuple, [...def]];
}
function unionF(def) {
  return [URI.union, [...def]];
}
function intersectF(def) {
  return [URI.intersect, [...def]];
}
const NullaryTags = [URI.never, URI.any, URI.unknown, URI.void, URI.undefined, URI.null, URI.boolean, URI.symbol];
const isNullary = u => NullaryTags.includes(u);
export const laxMin = (...xs) => {
  const ys = xs.filter(t.number);
  return ys.length === 0 ? void 0 : Math.min(...ys);
};
export const laxMax = (...xs) => {
  const ys = xs.filter(t.number);
  return ys.length === 0 ? void 0 : Math.max(...ys);
};
const makeInclusiveBounds = model => ({
  minimum: model,
  maximum: model
});
const makeExclusiveBounds = model => ({
  exclusiveMinimum: model,
  exclusiveMaximum: model
});
/**
 * It's _crazy_ how complicated this is.
 *
 * Surely there's a more elegant way to accomplish this.
 */
export const preprocessNumberBounds = ({
  minimum: min_,
  maximum: max_,
  exclusiveMinimum: xMin_,
  exclusiveMaximum: xMax_
}) => {
  let min = laxMin(min_, max_);
  let max = laxMax(min_, max_);
  let xMin = laxMin(xMin_, xMax_);
  let xMax = laxMax(xMin_, xMax_);
  let noExclusiveMin = xMin === xMax && xMin !== xMin_;
  let noExclusiveMax = xMin === xMax && xMax !== xMax_ || t.number(xMin) && t.number(xMax) && xMin === xMax;
  let noMin = t.number(xMin) && !noExclusiveMin || min === max && min !== min_ || !t.number(min) || t.number(xMin) || t.number(xMax) && min > xMax;
  let noMax = t.number(xMax) && !noExclusiveMax || min === max && max !== max_ || !t.number(max) || t.number(xMax) || t.number(xMin) && max > xMin;
  return unsafeCompact({
    minimum: noMin ? void 0 : min,
    maximum: noMax ? void 0 : max,
    exclusiveMinimum: noExclusiveMin ? void 0 : xMin,
    exclusiveMaximum: noExclusiveMax ? void 0 : xMax
  });
};
const pickBounds = ({
  minimum,
  maximum
}) => {
  if (isNumeric(minimum)) {
    if (isNumeric(maximum)) return {
      minimum: minimum < maximum ? minimum : maximum,
      maximum: minimum > maximum ? minimum : maximum
    };else return {
      minimum
    };
  } else if (isNumeric(maximum)) return {
    maximum
  };else return void 0;
};
const stringBounds = /*#__PURE__*/fc.record(makeInclusiveBounds(fc.integer(defaultStringConstraints)), {
  requiredKeys: []
}).map(pickBounds);
const arrayBounds = /*#__PURE__*/fc.record(makeInclusiveBounds(fc.integer(defaultArrayConstraints)), {
  requiredKeys: []
}).map(pickBounds);
const integerBounds = /*#__PURE__*/fc.record(makeInclusiveBounds(fc.integer(defaultIntegerConstraints)), {
  requiredKeys: []
}).map(pickBounds);
const bigintBounds = /*#__PURE__*/fc.record({
  ...makeInclusiveBounds(fc.bigInt(defaultBigIntConstraints)),
  ...makeExclusiveBounds(fc.bigInt(defaultBigIntConstraints))
}, {
  requiredKeys: []
}).map(pickBounds);
const numberBounds = /*#__PURE__*/fc.record({
  ...makeInclusiveBounds(fc.float(defaultNumberConstraints)),
  ...makeExclusiveBounds(fc.float(defaultNumberConstraints))
}, {
  requiredKeys: []
}).map(preprocessNumberBounds);
const BoundableTags = [URI.bigint, URI.integer, URI.number, URI.string];
const isBoundableTag = u => BoundableTags.includes(u);
const isBoundable = u => t.has(0, isBoundableTag)(u);
const SpecialCaseTags = [URI.eq];
const isSpecialCaseTag = u => SpecialCaseTags.includes(u);
const isSpecialCase = u => Array_isArray(u) && isSpecialCaseTag(u[0]);
const UnaryTags = [URI.array, URI.record, URI.optional];
const isUnaryTag = u => UnaryTags.includes(u);
const isUnary = u => Array_isArray(u) && isUnaryTag(u[0]);
const PositionalTags = [URI.union, URI.intersect, URI.tuple];
const isPositionalTag = u => PositionalTags.includes(u);
const isPositional = u => Array_isArray(u) && isPositionalTag(u[0]) && Array_isArray(u[1]);
const AssociativeTags = [URI.object];
const isAssociativeTag = u => AssociativeTags.includes(u);
const isAssociative = u => Array_isArray(u) && isAssociativeTag(u[0]);
const isSeed = u => isNullary(u) || isUnary(u) || isBoundable(u) || isPositional(u) || isAssociative(u) || isSpecialCase(u);
/**
 * ## {@link is `is`}
 *
 * Type-guard from `unknown` to {@link Seed `Seed`}.
 *
 * The {@link is `is`} function is also an object whose properties narrow
 * to a particular member or subset of members of {@link Seed `Seed`}.
 */
function is(u) {
  return isSeed(u);
}
is.nullary = isNullary;
is.unary = isUnary;
is.positional = isPositional;
is.associative = isAssociative;
is.special = isSpecialCase;
is.never = u => u === URI.never;
is.any = u => u === URI.any;
is.unknown = u => u === URI.unknown;
is.void = u => u === URI.void;
is.null = u => u === URI.null;
is.undefined = u => u === URI.undefined;
is.boolean = u => u === URI.boolean;
is.symbol = u => u === URI.symbol;
is.string = /*#__PURE__*/t.has(0, u => u === URI.string);
is.number = /*#__PURE__*/t.has(0, u => u === URI.number);
is.integer = /*#__PURE__*/t.has(0, u => u === URI.integer);
is.bigint = /*#__PURE__*/t.has(0, u => u === URI.bigint);
is.eq = u => Array_isArray(u) && u[0] === URI.eq;
is.array = u => Array_isArray(u) && u[0] === URI.array;
is.optional = u => Array_isArray(u) && u[0] === URI.optional;
is.record = u => Array_isArray(u) && u[0] === URI.record;
is.union = u => Array_isArray(u) && u[0] === URI.union;
is.intersect = u => Array_isArray(u) && u[0] === URI.intersect;
is.tuple = u => Array_isArray(u) && u[0] === URI.tuple;
is.object = u => Array_isArray(u) && u[0] === URI.object;
/**
 * Hand-tuned constructor that gives you both more precise and
 * more localized feedback than the TS compiler when you make a mistake
 * in an inline {@link Seed `Seed`} definition.
 *
 * When you're working with deeply nested tuples where only certain sequences
 * are valid constructions, this turns out to be pretty useful / necessary.
 */
function defineSeed(seed) {
  return seed;
}
const NullarySchemaMap = {
  [URI.never]: t.never,
  [URI.void]: t.void,
  [URI.unknown]: t.unknown,
  [URI.any]: t.any,
  [URI.symbol]: t.symbol,
  [URI.null]: t.null,
  [URI.undefined]: t.undefined,
  [URI.boolean]: t.boolean
};
const BoundableSchemaMap = {
  [URI.integer]: bounds => {
    let schema = t.integer;
    if (!bounds) return schema;
    if (t.integer(bounds.minimum)) void (schema = schema.min(bounds.minimum));
    if (t.integer(bounds.maximum)) void (schema = schema.max(bounds.maximum));
    return schema;
  },
  [URI.bigint]: bounds => {
    let schema = t.bigint;
    if (!bounds) return schema;
    if (t.bigint(bounds.minimum)) void (schema = schema.min(bounds.minimum));
    if (t.bigint(bounds.maximum)) void (schema = schema.max(bounds.maximum));
    return schema;
  },
  [URI.number]: bounds => {
    let schema = t.number;
    if (!bounds) return schema;
    if (t.number(bounds.exclusiveMinimum)) void (schema = schema.moreThan(bounds.exclusiveMinimum));
    if (t.number(bounds.exclusiveMaximum)) void (schema = schema.lessThan(bounds.exclusiveMaximum));
    if (t.number(bounds.minimum)) void (schema = schema.min(bounds.minimum));
    if (t.number(bounds.maximum)) void (schema = schema.max(bounds.maximum));
    return schema;
  },
  [URI.string]: bounds => {
    let schema = t.string;
    if (!bounds) return schema;
    if (t.integer(bounds.minimum)) void (schema = schema.min(bounds.minimum));
    if (t.integer(bounds.maximum)) void (schema = schema.max(bounds.maximum));
    return schema;
  },
  [URI.array]: (bounds, child) => {
    let schema = t.array.def(!child ? t.unknown : child);
    if (!bounds) return schema;
    if (t.integer(bounds?.minimum)) void (schema = schema.min(bounds.minimum));
    if (t.integer(bounds?.maximum)) void (schema = schema.max(bounds.maximum));
    return schema;
  }
};
const NullaryArbitraryMap = {
  [URI.never]: /*#__PURE__*/fc.constant(void 0),
  [URI.void]: /*#__PURE__*/fc.constant(void 0),
  [URI.unknown]: /*#__PURE__*/fc.jsonValue(),
  [URI.any]: /*#__PURE__*/fc.jsonValue(),
  [URI.symbol]: /*#__PURE__*/fc.string().map(Symbol.for),
  [URI.null]: /*#__PURE__*/fc.constant(null),
  [URI.undefined]: /*#__PURE__*/fc.constant(undefined),
  [URI.boolean]: /*#__PURE__*/fc.boolean()
};
const integerConstraintsFromBounds = (bounds = {}) => {
  const {
    maximum: max = NaN,
    minimum: min = NaN
  } = bounds;
  let constraints = {};
  let minimum = getMin(min, max);
  let maximum = getMax(max, min);
  if (t.integer(minimum)) constraints.min = minimum;
  if (t.integer(maximum)) constraints.max = maximum;
  return constraints;
};
const numberConstraintsFromBounds = (bounds = {}) => {
  if (Object.keys(bounds).length === 0) return {};
  let {
    minimum: min_,
    maximum: max_,
    exclusiveMinimum: xMin,
    exclusiveMaximum: xMax
  } = bounds;
  let exclusiveMinimum = isNumeric(xMin) ? isNumeric(xMax) ? getExclusiveMin(xMin, xMax) : xMin : void 0;
  let exclusiveMaximum = isNumeric(xMax) ? isNumeric(xMin) ? getExclusiveMax(xMax, xMin) : xMax : void 0;
  let minimum = isNumeric(min_) ? isNumeric(max_) ? getMin(min_, max_) : min_ : void 0;
  let maximum = isNumeric(max_) ? isNumeric(min_) ? getMax(max_, min_) : max_ : void 0;
  let min = isNumeric(exclusiveMinimum) ? !isNumeric(minimum) ? exclusiveMinimum : getMax(minimum, exclusiveMinimum) : minimum;
  let max = isNumeric(exclusiveMaximum) ? !isNumeric(maximum) ? exclusiveMaximum : getMin(maximum, exclusiveMaximum) : maximum;
  return unsafeCompact({
    min: isNumeric(min) ? getMin(min, max) : void 0,
    max: isNumeric(max) ? getMax(max, min) : void 0,
    minExcluded: isNumeric(exclusiveMinimum) && min === exclusiveMinimum,
    maxExcluded: isNumeric(exclusiveMaximum) && max === exclusiveMaximum
  });
};
const isNaN = globalThis.Number.isNaN;
const absorbNaN = x => isNaN(x) ? void 0 : x;
const getMin = (x_, y_, x = absorbNaN(x_), y = absorbNaN(y_)) => absorbNaN(x === void 0 ? void 0 : y === void 0 ? x : x <= y ? x : y <= x ? y : void 0);
const getExclusiveMin = (x_, y_, x = absorbNaN(x_), y = absorbNaN(y_)) => absorbNaN(x === void 0 ? void 0 : y === void 0 ? x : x < y ? x : y < x ? y : void 0);
const getMax = (x_, y_, x = absorbNaN(x_), y = absorbNaN(y_)) => absorbNaN(x === void 0 ? void 0 : y === void 0 ? x : x >= y ? x : y >= x ? y : void 0);
const getExclusiveMax = (x_, y_, x = absorbNaN(x_), y = absorbNaN(y_)) => absorbNaN(x === void 0 ? void 0 : y === void 0 ? x : x > y ? x : y > x ? y : void 0);
const bigintConstraintsFromBounds = ({
  maximum: max = NaN,
  minimum: min = NaN
} = {}) => {
  let constraints = {};
  let minimum = getMin(min, max);
  let maximum = getMax(max, min);
  if (t.bigint(minimum)) constraints.min = minimum;
  if (t.bigint(maximum)) constraints.max = maximum;
  return constraints;
};
const stringConstraintsFromBounds = ({
  minimum: min = NaN,
  maximum: max = NaN
} = {}) => {
  let constraints = {};
  let minimum = getMin(min, max);
  let maximum = getMax(max, min);
  if (t.integer(minimum)) void (constraints.minLength = minimum);
  if (t.integer(maximum)) if (maximum >= (constraints.minLength ?? Number.MIN_SAFE_INTEGER)) void (constraints.maxLength = maximum);
  return constraints;
};
const arrayConstraintsFromBounds = ({
  minimum: min = NaN,
  maximum: max = NaN
} = {}) => {
  let constraints = {};
  let minimum = getMin(min, max);
  let maximum = getMax(max, min);
  if (t.integer(minimum)) void (constraints.minLength = minimum);
  if (t.integer(maximum)) if (maximum >= (constraints.minLength ?? Number.MIN_SAFE_INTEGER)) void (constraints.maxLength = maximum);
  return constraints;
};
const BoundableArbitraryMap = {
  [URI.integer]: bounds => fc.integer(integerConstraintsFromBounds(bounds)),
  [URI.number]: bounds => fc.float(numberConstraintsFromBounds(bounds)),
  [URI.bigint]: bounds => fc.bigInt(bigintConstraintsFromBounds(bounds)),
  [URI.string]: bounds => fc.string(stringConstraintsFromBounds(bounds)),
  [URI.array]: (arb, bounds) => fc.array(arb, arrayConstraintsFromBounds(bounds))
};
const NullaryStringMap = {
  [URI.never]: 'never',
  [URI.void]: 'void',
  [URI.unknown]: 'unknown',
  [URI.any]: 'any',
  [URI.symbol]: 'symbol',
  [URI.null]: 'null',
  [URI.undefined]: 'undefined',
  [URI.boolean]: 'boolean'
};
const BoundableStringMap = {
  [URI.integer]: 'integer',
  [URI.number]: 'number',
  [URI.bigint]: 'bigint',
  [URI.string]: 'string',
  [URI.array]: 'array'
};
const BoundableSeedMap = {
  [URI.integer]: integerF,
  [URI.number]: numberF,
  [URI.bigint]: bigintF,
  [URI.string]: stringF,
  [URI.array]: arrayF
};
const Functor = {
  map(f) {
    return x => {
      if (!isSeed(x)) return x;
      switch (true) {
        default:
          return x;
        case isBoundable(x):
          return BoundableSeedMap[x[0]](x[1]);
        case isNullary(x):
          return x;
        case x[0] === URI.eq:
          return eqF(x[1]);
        case x[0] === URI.array:
          return arrayF(f(x[1]), x[2]);
        case x[0] === URI.record:
          return recordF(f(x[1]));
        case x[0] === URI.optional:
          return optionalF(f(x[1]));
        case x[0] === URI.tuple:
          return tupleF(x[1].map(f));
        case x[0] === URI.union:
          return unionF(x[1].map(f));
        case x[0] === URI.intersect:
          return intersectF(x[1].map(f));
        case x[0] === URI.object:
          return objectF(x[1].map(([k, v]) => [k, f(v)]));
      }
    };
  }
};
const fold = /*#__PURE__*/fn.cata(Functor);
const unfold = /*#__PURE__*/fn.ana(Functor);
const normalizeMin = x => t.has('minimum', t.union(t.number, t.bigint))(x) ? x.minimum : t.has('minLength')(x) ? x.minLength : void 0;
const normalizeMax = x => t.has('maximum', t.union(t.number, t.bigint))(x) ? x.maximum : t.has('maxLength')(x) ? x.maxLength : void 0;
const normalizeExclusiveMin = x => t.has('exclusiveMinimum', isNumeric)(x) ? x.exclusiveMinimum : void 0;
const normalizeExclusiveMax = x => t.has('exclusiveMaximum', isNumeric)(x) ? x.exclusiveMaximum : void 0;
function getBounds(x) {
  let min_ = normalizeMin(x);
  let max_ = normalizeMax(x);
  let xMin_ = normalizeExclusiveMin(x);
  let xMax_ = normalizeExclusiveMax(x);
  let min = isNumeric(min_) ? getMin(min_, max_) : void 0;
  let max = isNumeric(max_) ? getMax(max_, min_) : void 0;
  let xMin = isNumeric(xMin_) ? getMin(xMin_, xMax_) : void 0;
  let xMax = isNumeric(xMax_) ? getMax(xMax_, xMin_) : void 0;
  let out = unsafeCompact({
    exclusiveMinimum: isNumeric(xMin) ? xMin : void 0,
    exclusiveMaximum: isNumeric(xMax) ? xMax : void 0,
    minimum: isNumeric(xMin) ? void 0 : isNumeric(min) ? isNumeric(xMax) ? xMax < min ? void 0 : min : min : void 0,
    maximum: isNumeric(xMax) ? void 0 : isNumeric(max) ? isNumeric(xMin) ? max < xMin ? void 0 : max : max : void 0
  });
  return Object.keys(out).length === 0 ? void 0 : out;
}
var Recursive;
(function (Recursive) {
  Recursive.identity = x => x;
  // export const sort: T.Functor.Coalgebra<Seed.Free, Seed.Fixpoint> = (x) =>
  //   typeof x !== 'string' && x[0] === URI.tuple
  //     ? [x[0], [...x[1]].sort(sortSeedOptionalsLast)]
  //     : x
  Recursive.toExtensibleSchema = $ => x => {
    if (!isSeed(x)) return fn.exhaustive(x);
    switch (true) {
      default:
        return fn.exhaustive(x);
      case isNullary(x):
        return NullarySchemaMap[x];
      case x[0] === URI.never:
        return $?.never ?? NullarySchemaMap[x];
      case x[0] === URI.unknown:
        return $?.unknown ?? NullarySchemaMap[x];
      case x[0] === URI.void:
        return $?.void ?? NullarySchemaMap[x];
      case x[0] === URI.any:
        return $?.any ?? NullarySchemaMap[x];
      case x[0] === URI.undefined:
        return $?.undefined ?? NullarySchemaMap[x];
      case x[0] === URI.null:
        return $?.null ?? NullarySchemaMap[x];
      case x[0] === URI.symbol:
        return $?.symbol ?? NullarySchemaMap[x];
      case x[0] === URI.boolean:
        return $?.boolean ?? NullarySchemaMap[x];
      case x[0] === URI.integer:
        return $?.integer ?? BoundableSchemaMap[x[0]](x[1]);
      case x[0] === URI.bigint:
        return $?.bigint ?? BoundableSchemaMap[x[0]](x[1]);
      case x[0] === URI.number:
        return $?.number ?? BoundableSchemaMap[x[0]](x[1]);
      case x[0] === URI.string:
        return $?.string ?? BoundableSchemaMap[x[0]](x[1]);
      case x[0] === URI.array:
        return $?.array?.(x[1]) ?? BoundableSchemaMap[x[0]](x[2], x[1]);
      case x[0] === URI.eq:
        return $?.eq?.(x[1]) ?? t.eq.def(x[1]);
      case x[0] === URI.record:
        return $?.record?.(x[1]) ?? t.record.def(x[1]);
      case x[0] === URI.optional:
        return $?.optional?.(x[1]) ?? t.optional.def(x[1]);
      case x[0] === URI.tuple:
        return $?.tuple?.(x[1]) ?? t.tuple.def([...x[1]].sort(sortOptionalsLast), opts);
      case x[0] === URI.union:
        return $?.union?.(x[1]) ?? t.union.def(x[1]);
      case x[0] === URI.intersect:
        return $?.intersect?.(x[1]) ?? t.intersect.def(x[1]);
      case x[0] === URI.object:
        {
          const wrap = $?.object ?? t.object;
          return wrap(Object_fromEntries(x[1].map(([k, v]) => [parseKey(k), v])), opts);
        }
    }
  };
  Recursive.toSchema = x => {
    if (!isSeed(x)) return x; // fn.exhaustive(x)
    switch (true) {
      default:
        return fn.exhaustive(x);
      case isNullary(x):
        return NullarySchemaMap[x];
      case x[0] === URI.array:
        return BoundableSchemaMap[x[0]](x[2], x[1]);
      case isBoundable(x):
        return BoundableSchemaMap[x[0]](x[1]);
      case x[0] === URI.eq:
        return t.eq.def(x[1]);
      case x[0] === URI.record:
        return t.record.def(x[1]);
      case x[0] === URI.optional:
        return t.optional.def(x[1]);
      case x[0] === URI.tuple:
        return t.tuple.def([...x[1]].sort(sortOptionalsLast), opts);
      case x[0] === URI.union:
        return t.union.def(x[1]);
      case x[0] === URI.intersect:
        return t.intersect.def(x[1]);
      case x[0] === URI.object:
        return t.object.def(Object_fromEntries(x[1].map(([k, v]) => [parseKey(k), v])), opts);
    }
  };
  Recursive.fromSchema = x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case t.isNullary(x):
        return x.tag;
      case t.isBoundable(x):
        return BoundableSeedMap[x.tag](getBounds(x));
      case x.tag === URI.array:
        return arrayF(x.def, getBounds(x));
      case x.tag === URI.record:
        return [x.tag, x.def];
      case x.tag === URI.optional:
        return [x.tag, x.def];
      case x.tag === URI.eq:
        return [x.tag, x.def];
      case x.tag === URI.tuple:
        return [x.tag, x.def];
      case x.tag === URI.union:
        return [x.tag, x.def];
      case x.tag === URI.intersect:
        return [x.tag, x.def];
      case x.tag === URI.object:
        return [x.tag, Object.entries(x.def)];
    }
  };
  Recursive.toArbitrary = x => {
    if (!isSeed(x)) return fn.exhaustive(x);
    switch (true) {
      default:
        return fn.exhaustive(x);
      case isNullary(x):
        return NullaryArbitraryMap[x];
      case isBoundable(x):
        return BoundableArbitraryMap[x[0]](x[1]);
      case x[0] === URI.eq:
        return fc.constant(x[1]);
      case x[0] === URI.array:
        return BoundableArbitraryMap[x[0]](x[1], x[2]);
      case x[0] === URI.record:
        return fc.dictionary(fc.identifier(), x[1]);
      case x[0] === URI.optional:
        return fc.optional(x[1]);
      case x[0] === URI.tuple:
        return fc.tuple(...x[1]);
      case x[0] === URI.union:
        return fc.oneof(...x[1]);
      case x[0] === URI.object:
        {
          const out = fc.record(Object_fromEntries(x[1]));
          return out;
        }
      case x[0] === URI.intersect:
        {
          if (x[1].length === 1) return x[1][0];
          const ys = x[1].filter(_ => !isNullary(_));
          return ys.length === 0 ? x[1][0] : fc.tuple(...ys).map(([head, ...tail]) => !isComposite(head) ? head : tail.reduce((acc, y) => isComposite(y) ? Object_assign(acc, y) : acc, head));
        }
    }
  };
  Recursive.toString = x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case isNullary(x):
        return NullaryStringMap[x];
      case isBoundable(x):
        return BoundableStringMap[x[0]];
      case x[0] === URI.eq:
        return x[1];
      case x[0] === URI.array:
        return 'Array<' + x[1] + '>';
      case x[0] === URI.record:
        return 'Record<string, ' + x[1] + '>';
      case x[0] === URI.optional:
        return x[1] + '?';
      case x[0] === URI.tuple:
        return '[' + x[1].join(', ') + ']';
      case x[0] === URI.union:
        return x[1].join(' | ');
      case x[0] === URI.intersect:
        return x[1].join(' & ');
      case x[0] == URI.object:
        return '{ ' + x[1].flatMap(([k, v]) => `${parseKey(k)}: ${v}`).join(', ') + ' }';
    }
  };
  Recursive.fromJsonLiteral = x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case x === null:
        return URI.null;
      case x === void 0:
        return URI.undefined;
      case typeof x === 'boolean':
        return URI.boolean;
      case typeof x === 'number':
        return [URI.number, {}];
      case typeof x === 'string':
        return [URI.string, {}];
      case Json.isArray(x):
        return [URI.tuple, x];
      case Json.isObject(x):
        return [URI.object, Object.entries(x)];
    }
  };
})(Recursive || (Recursive = {}));
const hasOptionalTag = /*#__PURE__*/t.has('tag', x => x === URI.optional);
export const sortOptionalsLast = (l, r) => hasOptionalTag(l) ? 1 : hasOptionalTag(r) ? -1 : 0;
const sortSeedOptionalsLast = (l, r) => isOptional(l) ? 1 : isOptional(r) ? -1 : 0;
const isOptional = node => typeof node === 'string' ? false : node[0] === URI.optional;
export const pickAndSortNodes = nodes => ({
  include,
  exclude,
  sortBias
} = defaults) => {
  const sortFn = sortBias === undefined ? defaults.sortBias : typeof sortBias === 'function' ? sortBias : (l, r) => (sortBias[l] ?? 0) < (sortBias[r] ?? 0) ? 1 : (sortBias[l] ?? 0) > (sortBias[r] ?? 0) ? -1 : 0;
  return nodes.filter(x => (include ? include.includes(x) : true) && (exclude ? !exclude.includes(x) : true)).sort(sortFn);
};
function parseConstraints({
  exclude = defaults.exclude,
  include = defaults.include,
  sortBias = defaults.sortBias,
  // groupScalars = defaults.groupScalars,
  integer: {
    min: integerMin = defaults.integer.min,
    max: integerMax = defaults.integer.max
  } = defaults.integer,
  bigint: {
    min: bigintMin = defaults.bigint.min,
    max: bigintMax = defaults.bigint.max
  } = defaults.bigint,
  number: {
    min: numberMin = defaults.number.min,
    max: numberMax = defaults.number.max,
    minExcluded: numberMinExcluded = defaults.number.minExcluded,
    maxExcluded: numberMaxExcluded = defaults.number.maxExcluded,
    ...numberRest
  } = defaults.number,
  string: {
    min: stringMinLength = defaults.string.min,
    max: stringMaxLength = defaults.string.max,
    ...stringRest
  } = defaults.string,
  array: {
    min: arrayMinLength = defaults.array.min,
    max: arrayMaxLength = defaults.array.max
  } = defaults.array,
  intersect: {
    minLength: intersectMinLength,
    maxLength: intersectMaxLength,
    size: intersectSize,
    depthIdentifier: intersectDepthIdentifier
  } = defaults.intersect,
  union: {
    minLength: unionMinLength,
    maxLength: unionMaxLength,
    size: unionSize
  } = defaults.union,
  tuple: {
    minLength: tupleMinLength = defaults.tuple.minLength,
    maxLength: tupleMaxLength = defaults.tuple.maxLength,
    size: tupleSize = defaults.tuple.size,
    depthIdentifier: tupleDepthIdentifier = defaults.tuple.depthIdentifier
  } = defaults.tuple,
  object: {
    min: objectMinLength = defaults.object.min,
    max: objectMaxLength = defaults.object.max,
    size: objectSize = defaults.object.size
  } = defaults.object,
  tree: {
    depthIdentifier: treeDepthIdentifier = defaults.tree.depthIdentifier,
    depthSize: treeDepthSize = defaults.tree.depthSize,
    maxDepth: treeMaxDepth = defaults.tree.maxDepth,
    withCrossShrink: treeWithCrossShrink = defaults.tree.withCrossShrink
  } = defaults.tree
} = defaults) {
  const integer = {
    min: integerMin,
    max: integerMax
  };
  const bigint = {
    min: bigintMin,
    max: bigintMax
  };
  const number = {
    min: numberMin,
    max: numberMax,
    minExcluded: numberMinExcluded,
    maxExcluded: numberMaxExcluded,
    ...numberRest
  };
  const string = {
    min: stringMinLength,
    max: stringMaxLength,
    ...stringRest
  };
  const array = {
    min: arrayMinLength,
    max: arrayMaxLength
  };
  const object = {
    size: objectSize,
    minLength: objectMinLength,
    maxLength: objectMaxLength
  };
  const tree = {
    depthIdentifier: treeDepthIdentifier,
    depthSize: treeDepthSize,
    maxDepth: treeMaxDepth,
    withCrossShrink: treeWithCrossShrink
  };
  const tuple = {
    depthIdentifier: tupleDepthIdentifier,
    minLength: tupleMinLength,
    maxLength: tupleMaxLength,
    size: tupleSize
  };
  const intersect = {
    depthIdentifier: intersectDepthIdentifier,
    minLength: intersectMinLength,
    maxLength: intersectMaxLength,
    size: intersectSize
  };
  const union = {
    depthIdentifier: defaultDepthIdentifier,
    minLength: unionMinLength,
    maxLength: unionMaxLength,
    size: unionSize
  };
  return {
    exclude: exclude,
    include: include.filter(_ => !exclude.includes(_)),
    integer,
    bigint,
    number,
    string,
    array,
    intersect,
    object,
    sortBias,
    tree,
    tuple,
    union
    // groupScalars,
  };
}
const NullaryJsonMap = {
  [URI.never]: void 0,
  [URI.unknown]: void 0,
  [URI.void]: void 0,
  [URI.any]: void 0,
  [URI.undefined]: void 0,
  [URI.null]: null,
  [URI.symbol]: /*#__PURE__*/globalThis.Symbol().toString(),
  [URI.boolean]: false,
  [URI.bigint]: 0,
  [URI.integer]: 0,
  [URI.number]: 0,
  [URI.string]: ""
};
const isKeyOf = (k, x) => !!x && typeof x === 'object' && (typeof k === 'string' || typeof k === 'number' || typeof k === 'symbol') && k in x;
export const toJson = /*#__PURE__*/fold(x => {
  if (x == null) return x;
  switch (true) {
    default:
      return x;
    case isKeyOf(x, NullaryJsonMap):
      return NullaryJsonMap[x];
    case x[0] === URI.number:
      return 0;
    case x[0] === URI.integer:
      return 0;
    case x[0] === URI.bigint:
      return 0;
    case x[0] === URI.string:
      return '';
    case x[0] === URI.eq:
      return toJson(x[1]);
    case x[0] === URI.array:
      return [];
    case x[0] === URI.record:
      return {};
    case x[0] === URI.optional:
      return x[1];
    case x[0] === URI.object:
      return Object_fromEntries(x[1]);
    case x[0] === URI.tuple:
      return x[1];
    case x[0] === URI.record:
      return x[1];
    case x[0] === URI.union:
      return x[1][0];
    case x[0] === URI.intersect:
      return x[1].reduce((acc, y) => acc == null ? acc : y == null ? y : Object_assign(acc, y), {});
  }
});
const Nullaries = {
  never: /*#__PURE__*/fc.constant(URI.never),
  any: /*#__PURE__*/fc.constant(URI.any),
  unknown: /*#__PURE__*/fc.constant(URI.unknown),
  void: /*#__PURE__*/fc.constant(URI.void),
  null: /*#__PURE__*/fc.constant(URI.null),
  undefined: /*#__PURE__*/fc.constant(URI.undefined),
  symbol: /*#__PURE__*/fc.constant(URI.symbol),
  boolean: /*#__PURE__*/fc.constant(URI.boolean)
};
const isEmpty = x => !x || Object.keys(x).length === 0;
const dropEmptyBounds = ([uri, bounds]) => isEmpty(bounds) ? [uri] : [uri, bounds];
const Boundables = {
  integer: /*#__PURE__*/fc.tuple(fc.constant(URI.integer), integerBounds).map(dropEmptyBounds),
  bigint: /*#__PURE__*/fc.tuple(fc.constant(URI.bigint), bigintBounds).map(dropEmptyBounds),
  number: /*#__PURE__*/fc.tuple(fc.constant(URI.number), numberBounds).map(dropEmptyBounds),
  string: /*#__PURE__*/fc.tuple(fc.constant(URI.string), stringBounds).map(dropEmptyBounds)
};
const Unaries = {
  eq: (fix, _) => fix.chain(() => fc.jsonValue()).map(eqF),
  array: (fix, $) => fc.tuple(fix, arrayBounds).map(([def, bounds]) => arrayF(def, bounds)),
  record: (fix, _) => fix.map(recordF),
  optional: (fix, _) => fc.optional(fix).map(optionalF),
  tuple: (fix, $) => fc.array(fix, $.tuple).map(fn.flow(_ => _.sort(sortSeedOptionalsLast), tupleF)),
  object: (fix, $) => fc.entries(fix, $.object).map(objectF),
  union: (fix, $) => fc.array(fix, $.union).map(unionF),
  intersect: (fix, $) => fc.array(fix, $.intersect).map(intersectF)
};
function getNullaries(typeNames) {
  return Object.fromEntries(Object.keys(Nullaries).filter(nullary => typeNames.includes(nullary)).map(nullary => [nullary, Nullaries[nullary]]));
}
function getBoundables(typeNames) {
  return Object.fromEntries(Object.keys(Boundables).filter(boundable => typeNames.includes(boundable)).map(boundable => [boundable, Boundables[boundable]]));
}
function getUnaries(typeNames, $, fix) {
  return Object.fromEntries(Object.keys(Unaries).filter(unary => typeNames.includes(unary)).map(unary => [unary, Unaries[unary](fix, $)]));
}
function seed(_ = defaults) {
  const $ = parseConstraints(_);
  const nodes = pickAndSortNodes(initialOrder)($);
  return go => {
    const builder = {
      ...getNullaries(nodes),
      ...getBoundables(nodes),
      ...getUnaries(nodes, $, go('tree'))
    };
    return {
      ...builder,
      tree: fc.oneof($.tree, ...Object.values(builder))
    };
  };
}
function seedWithChain(_ = defaults) {
  const $ = parseConstraints(_);
  const nodes = pickAndSortNodes(initialOrder)($);
  return go => {
    const builder = {
      ...getNullaries(nodes),
      ...getBoundables(nodes),
      ...getUnaries(nodes, $, go('tree'))
    };
    return {
      ...builder,
      tree: fc.oneof($.tree, ...Object.values(builder))
    };
  };
}
const minDepth = {
  array: (seeds, _) => fc.oneof(...seeds).map(arrayF),
  record: (seeds, _) => fc.oneof(...seeds).map(recordF),
  optional: (seeds, $) => fc.oneof(...seeds).map(optionalF),
  object: (seeds, $) => fc.array(fc.tuple(fc.identifier(), fc.oneof(...seeds)), {
    maxLength: $.object.maxLength,
    minLength: $.object.minLength
  }).map(objectF),
  tuple: (seeds, $) => fc.array(fc.oneof(...seeds), {
    minLength: $.tuple.minLength,
    maxLength: $.tuple.maxLength
  }).map(tupleF),
  union: (seeds, $) => fc.array(fc.oneof(...seeds), {
    minLength: $.union.minLength,
    maxLength: $.union.maxLength
  }).map(unionF),
  intersect: (seeds, $) => fc.array(fc.oneof(...seeds), {
    minLength: $.intersect.minLength,
    maxLength: $.intersect.maxLength
  }).map(intersectF)
};
const minDepthBranchOrder = ['object', 'optional', 'tuple', 'union', 'intersect', 'array', 'record'];
const minDepths = {
  [0]: minDepth[minDepthBranchOrder[0]],
  [1]: minDepth[minDepthBranchOrder[1]],
  [2]: minDepth[minDepthBranchOrder[2]],
  [3]: minDepth[minDepthBranchOrder[3]],
  [4]: minDepth[minDepthBranchOrder[4]],
  [5]: minDepth[minDepthBranchOrder[5]],
  [6]: minDepth[minDepthBranchOrder[6]]
};
function schemaWithMinDepth(_ = defaults, n) {
  let $ = parseConstraints(_);
  let seed = fc.letrec(seedWithChain($));
  let seeds = Object.values(seed);
  let branches = minDepthBranchOrder.filter(_ => $.include.includes(_) && !$.exclude.includes(_));
  let arb = seed.tree;
  while (n-- >= 0) arb = fc.nat(branches.length - 1).chain(x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case x === 0:
        return minDepths[x](seeds, $);
      case x === 1:
        return minDepths[x](seeds, $);
      case x === 2:
        return minDepths[x](seeds, $);
      case x === 3:
        return minDepths[x](seeds, $);
      case x === 4:
        return minDepths[x](seeds, $);
      case x === 5:
        return minDepths[x](seeds, $);
      case x === 6:
        return minDepths[x](seeds, $);
    }
  });
  return arb.map(toSchema);
}
const identity = /*#__PURE__*/fold(Recursive.identity);
//    ^?
const toSchema = /*#__PURE__*/fold(Recursive.toSchema);
//    ^?
const toArbitrary = /*#__PURE__*/fold(Recursive.toArbitrary);
//    ^?
const toString = /*#__PURE__*/fold(Recursive.toString);
//    ^?
const fromSchema = /*#__PURE__*/fn.cata(t.Functor)(Recursive.fromSchema);
//    ^?
const fromJsonLiteral = /*#__PURE__*/fold(Recursive.fromJsonLiteral);
function schema(constraints) {
  return fc.letrec(seed(constraints)).tree.map(toSchema);
}
const extensibleArbitrary = constraints => fc.letrec(seed(constraints)).tree.map(fold(Recursive.toExtensibleSchema(constraints?.arbitraries)));
/**
 * ## {@link data `Seed.data`}
 *
 * Generates an arbitrary,
 * [pseudo-random](https://en.wikipedia.org/wiki/Pseudorandomness)
 * data generator.
 *
 * Internally, schemas are generated from a
 * {@link Seed `seed value`}, which is itself generated by a library
 * called [`fast-check`](https://github.com/dubzzz/fast-check).
 */
const data = constraints => fc.letrec(seed(constraints)).tree.chain(toArbitrary);
//# sourceMappingURL=seed.js.map